from abc import abstractmethod
from typing import Any, Dict, Type

from selenium.webdriver.chromium.options import ChromiumOptions
from selenium.webdriver.chromium.service import ChromiumService

from pytest_logikal.browser.base import Browser

DEFAULT_JS_RANDOM_SEED = 42


class ChromiumBrowser(Browser):
    @property
    @abstractmethod
    def options_class(self) -> Type[ChromiumOptions]:
        ...

    @property
    @abstractmethod
    def service_class(self) -> Type[ChromiumService]:
        ...

    # See https://www.selenium.dev/documentation/webdriver/browsers/chrome/#options
    # See https://github.com/GoogleChrome/chrome-launcher/blob/main/docs/chrome-flags-for-tools.md
    def init_args(self) -> Dict[str, Any]:
        args = [
            '--headless',
            f'--window-size={self.settings.width},{self.settings.height}',
            # Unwanted features
            '--no-default-browser-check',
            '--no-first-run',
            '--ash-no-nudges',
            '--disable-search-engine-choice-screen',
            # Web platform behavior
            f'--js-flags=--random-seed={DEFAULT_JS_RANDOM_SEED}',
        ]

        options = self.options_class()
        options.binary_location = str(self.version.path)
        if self.settings.mobile:
            args.append('--hide-scrollbars')
        for arg in args:
            options.add_argument(arg)

        service = self.service_class(executable_path=str(self.version.driver_path))

        return {'options': options, 'service': service}



from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.webdriver import WebDriver

from pytest_logikal.browser.base import BrowserVersion
from pytest_logikal.browser.chromium import ChromiumBrowser
from pytest_logikal.browser.utils import download, make_executable, move, unzip
from pytest_logikal.utils import tmp_path


class Chrome(ChromiumBrowser, WebDriver):
    """
    Google Chrome WebDriver.
    """
    options_class = Options
    service_class = Service


class ChromeVersion(BrowserVersion):
    name = 'chrome'
    driver_class = Chrome

    def install(self) -> None:
        root = f'https://storage.googleapis.com/chrome-for-testing-public/{self.version}/linux64'

        if not self.path.exists():
            self.print(f'Installing Google Chrome {self.version}')
            tmp = tmp_path(self.name)
            unzip(download(f'{root}/chrome-linux64.zip', tmp / 'chrome.zip'))
            move(tmp / 'chrome/chrome-linux64', self.path.parent)
            make_executable(self.path)

        if not self.driver_path.exists():
            self.print(f'Installing Google Chrome WebDriver {self.driver_version}')
            tmp = tmp_path(self.driver_name)
            unzip(download(f'{root}/chromedriver-linux64.zip', tmp / 'chromedriver.zip'))
            move(tmp / 'chromedriver/chromedriver-linux64', self.driver_path.parent)
            make_executable(self.driver_path)



from dataclasses import dataclass
from typing import Iterable, Optional, Union


@dataclass
class Settings:
    """
    Browser settings specification.

    Args:
        name: Name of the settings.
        width: Browser window width.
        height: Browser window height.
        full_page_height: Whether to use the full page height for screenshots.
        mobile: Whether it is a mobile browser.

    """
    name: str
    width: int
    height: int
    full_page_height: bool = True
    mobile: bool = False


@dataclass
class Scenario:
    """
    Browser scenario specification.

    Args:
        settings: Settings to use.
        browsers: Browsers to use. Defaults to using all registered browser versions.

    """
    settings: Union[Settings, Iterable[Settings]]
    browsers: Optional[Iterable[str]] = None


desktop_4k = Scenario(Settings('desktop_4k', width=2560, height=1440))  #:
desktop = Scenario(Settings(
    'desktop',
    width=1920 - 120,  # offset for e.g. top bar, bottom bar, dash bar, task bar
    height=1080 - 180,  # offset for e.g. dash bar
))  #:
laptop_l = Scenario(Settings('laptop_l', width=1440, height=900))  #:
laptop = Scenario(Settings('laptop', width=1024, height=768))  #:
tablet = Scenario(Settings('tablet', width=768, height=1024, mobile=True))  #:
mobile_l = Scenario(Settings('mobile_l', width=425, height=680, mobile=True))  #:
mobile_m = Scenario(Settings('mobile_m', width=375, height=600, mobile=True))  #:
mobile_s = Scenario(Settings('mobile_s', width=320, height=512, mobile=True))  #:



# pylint: disable=import-outside-toplevel
import logging
from abc import ABC, abstractmethod
from contextlib import contextmanager
from importlib import import_module
from pathlib import Path
from time import sleep
from typing import Any, Dict, Iterator, Optional, Type

from selenium.webdriver.remote.webdriver import WebDriver
from termcolor import colored
from xdg import xdg_cache_home

from pytest_logikal.browser.scenarios import Settings
from pytest_logikal.utils import assert_image_equal, hide_traceback, tmp_path

logger = logging.getLogger(__name__)


class BrowserVersion(ABC):
    created: Dict[str, 'BrowserVersion'] = {}

    _initial_info = False

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def driver_class(self) -> Type['Browser']:
        ...

    @property
    def binary_name(self) -> str:
        return self.name

    @property
    def driver_name(self) -> str:
        return f'{self.name}driver'

    @property
    def driver_binary_name(self) -> str:
        return self.driver_name

    @property
    def driver_version(self) -> str:
        return self.version

    def __init__(self, version: str, install: bool = True, install_path: Optional[Path] = None):
        install_path = install_path or (xdg_cache_home() / 'pytest_logikal')

        self.version = version
        self.path = install_path / self.name / version / self.binary_name
        self.driver_path = install_path / self.driver_name / version / self.driver_binary_name
        if install:
            self.install()

        BrowserVersion.created[self.name] = self

    def __str__(self) -> str:
        return f'{self.name} ({self.version})'

    def __repr__(self) -> str:
        return (
            f'<{str(self.__class__.__name__)} ({self.version}) at "{self.path}" '
            f'with {self.driver_name} ({self.driver_version}) at "{self.driver_path}">'
        )

    @abstractmethod
    def install(self) -> None:
        ...

    def driver(self, **kwargs: Any) -> 'Browser':
        return self.driver_class(version=self, **kwargs)

    @staticmethod
    def print(message: str) -> None:
        if not BrowserVersion._initial_info:
            print(colored('Installing browsers', 'yellow', attrs=['bold']))
            BrowserVersion._initial_info = True
        print(f'\n{colored(message, attrs=["bold"])}')

    @staticmethod
    def final_info() -> None:
        if BrowserVersion._initial_info:
            print()  # trailing newline


class Browser(ABC, WebDriver):
    """
    Base class for browser-specific web drivers.
    """
    def __init__(
        self,
        *,
        version: BrowserVersion,
        settings: Settings,
        screenshot_path: Path = Path('screenshot'),
        image_tmp_path: Optional[Path] = None,
        **kwargs: Any,
    ):
        self.version = version
        logger.debug(f'Browser version: {self.version}')
        self.settings = settings
        logger.debug(f'Browser settings: {self.settings}')
        self.screenshot_path = screenshot_path
        logger.debug(f'Using screenshot path "{self.screenshot_path}"')
        self.image_tmp_path = image_tmp_path or tmp_path('browser')
        logger.debug(f'Using temporary path "{self.image_tmp_path}"')

        super().__init__(**{**kwargs, **self.init_args()})

    @abstractmethod
    def init_args(self) -> Dict[str, Any]:
        ...

    @contextmanager
    def auto_height(self, wait_milliseconds: Optional[int]) -> Iterator[None]:
        if not self.settings.full_page_height:
            yield
            return
        logger.debug('Using full page height')
        original_size = self.get_window_size()
        if wait_milliseconds:  # we use a small delay to mitigate height flakiness
            logger.debug(f'Waiting {wait_milliseconds} ms')
            sleep(wait_milliseconds / 1000)
        elements = [
            'document.body.clientHeight',
            'document.body.scrollHeight',
            'document.body.offsetHeight',
            'document.documentElement.clientHeight',
            'document.documentElement.scrollHeight',
            'document.documentElement.offsetHeight',
        ]
        script = f'return Math.max({",".join(elements)});'
        height = self.execute_script(script)  # type: ignore[no-untyped-call]
        logger.debug(f'Calculated page height: {height}')
        self.set_window_size(original_size['width'], height)
        try:
            yield
        finally:
            self.set_window_size(**original_size)

    @hide_traceback
    def check(self, name: Optional[str] = None, wait_milliseconds: Optional[int] = 100) -> None:
        """
        Create a screenshot and check it against an expected version.

        Args:
            name: The name of the expected screenshot.
            wait_milliseconds: The milliseconds to wait before calculating the screenshot height
                for unlimited height checks.

        """
        name_parts = [self.screenshot_path.name, self.version.name, self.settings.name, name]
        full_name = '_'.join(part for part in name_parts if part is not None)
        expected = self.screenshot_path.with_name(full_name).with_suffix('.png')

        script = 'document.body.style.caretColor = "transparent";'  # hide the blinking caret
        self.execute_script(script)  # type: ignore[no-untyped-call]

        with self.auto_height(wait_milliseconds=wait_milliseconds):
            logger.debug('Taking screenshot')
            # Note: we are disabling debug remote logs because they contain the verbose image data
            logging.getLogger('selenium.webdriver.remote').setLevel(logging.INFO)
            actual = self.get_screenshot_as_png()
            logging.getLogger('selenium.webdriver.remote').setLevel(logging.DEBUG)

        assert_image_equal(actual=actual, expected=expected, image_tmp_path=self.image_tmp_path)

    def login(self, user: Any, force: bool = True) -> None:
        """
        .. note:: The ``django`` extra must be installed for this method to work.

        Log in a given user.

        Args:
            user: The user to log in.
            force: Whether to log the user in without going through the authentication steps.

        """
        try:
            from django.conf import settings
            from django.contrib.auth import login as django_auth_login
            from django.http import HttpRequest
        except ImportError as error:  # pragma: no cover
            raise RuntimeError('The `django` extra must be installed for login to work') from error
        if not force:
            raise NotImplementedError('Only the forced login is implemented currently')

        request = HttpRequest()
        request.session = import_module(settings.SESSION_ENGINE).SessionStore()
        django_auth_login(request, user)
        request.session.save()
        self.add_cookie({
            'name': settings.SESSION_COOKIE_NAME,
            'value': request.session.session_key,
        })
